import base64
import enum
from dataclasses import dataclass
from datetime import date, datetime, time, timedelta, timezone
from decimal import Decimal
from typing import Any, Dict, FrozenSet, List, NamedTuple, Optional, Set, Tuple, TypedDict, cast

import attrs
import pyarrow as pa
import pytest
from pydantic import BaseModel

from chalk.features import DataFrame, feature, features, unwrap_feature
from chalk.features._encoding.serialized_dtype import (
    BaseSerializedDType,
    BinaryDType,
    DTypeCode,
    ListDType,
    SingletonDType,
    StructDType,
    StructField,
    TimestampDType,
    TimeUnitCode,
    TimeUnitDType,
)

try:
    import polars as pl
except ImportError:
    pytest.skip(reason="Polars not installed", allow_module_level=True)


class StrEnumFeature(enum.Enum):
    BLUE = "blue"
    GREEN = "green"
    ORANGE = "orange"


class IntEnumFeature(enum.IntEnum):
    ONE = 1
    TWO = 2
    THREE = 3


PL_STRUCT_SCHEMA = pl.Struct(
    [
        pl.Field("foo", pl.Int64),
        pl.Field("bar", pl.Utf8),
        pl.Field("baz", pl.Int64),
        pl.Field("sub_list", pl.List(pl.Int64)),
    ]
)

PA_STRUCT_SCHEMA = pa.struct(
    [
        pa.field("foo", pa.int64()),
        pa.field("bar", pa.utf8()),
        pa.field("baz", pa.int64()),
        pa.field("sub_list", pa.list_(pa.int64())),
    ]
)

SERIALIZED_STRUCT_DTYPE = StructDType(
    type_code=DTypeCode.STRUCT,
    fields=[
        StructField(name="foo", dtype=SingletonDType(type_code=DTypeCode.INT64)),
        StructField(name="bar", dtype=SingletonDType(type_code=DTypeCode.STRING)),
        StructField(name="baz", dtype=SingletonDType(type_code=DTypeCode.INT64)),
        StructField(
            name="sub_list",
            dtype=ListDType(type_code=DTypeCode.LIST, inner_dtype=SingletonDType(type_code=DTypeCode.INT64), length=-1),
        ),
    ],
)


class CustomClass:
    def __init__(self, foo: int, bar: str, baz: Optional[int], sub_list: List[int]):
        self.foo = foo
        self.bar = bar
        self.baz = baz
        self.sub_list = sub_list

    def __eq__(self, other: object):
        if not isinstance(other, CustomClass):
            return NotImplemented
        return (
            self.foo == other.foo
            and self.bar == other.bar
            and self.baz == other.baz
            and self.sub_list == other.sub_list
        )

    @classmethod
    def encoder(cls, instance: "CustomClass"):
        return instance.__dict__

    @classmethod
    def decoder(cls, x: Dict[str, Any]) -> "CustomClass":
        return CustomClass(**x)


@dataclass
class MyDataclass:
    foo: int
    bar: str
    baz: Optional[int]
    sub_list: List[int]


class MyNamedTuple(NamedTuple):
    foo: int
    bar: str
    baz: Optional[int]
    sub_list: List[int]


class MyTypedDict(TypedDict):
    foo: int
    bar: str
    baz: Optional[int]
    sub_list: List[int]


class PydanticFeature(BaseModel):
    foo: int
    bar: str
    baz: Optional[int]
    sub_list: List[int]


@attrs.define
class AttrsFeature:
    foo: int
    bar: str
    baz: Optional[int]
    sub_list: List[int]


class TypedDictFeature(TypedDict):
    foo: int
    bar: str
    baz: Optional[int]
    sub_list: List[int]


class SubStruct(TypedDict):
    name: str


class StructOfOptionalStructFeature(TypedDict):
    sub: Optional[SubStruct]


@features
class CodecFs:
    str_feature: str
    bool_feature: bool
    int_feature: int
    unsigned_int_feature: int = feature(dtype=pa.uint64())
    float_feature: float
    datetime_feature: datetime
    datetime_without_timezone_feature: datetime = feature(dtype=pa.timestamp("us"))
    time_feature: time
    date_feature: date
    duration_feature: timedelta
    binary_feature: bytes
    decimal_feature: Decimal
    str_enum_feature: StrEnumFeature
    int_enum_feature: IntEnumFeature
    dataclass_feature: MyDataclass
    named_tuple_feature: MyNamedTuple
    pydantic_feature: PydanticFeature
    attrs_feature: AttrsFeature
    typed_dict_feature: MyTypedDict
    custom_feature: CustomClass = feature(
        encoder=CustomClass.encoder,
        decoder=CustomClass.decoder,
        dtype=PA_STRUCT_SCHEMA,
    )
    list_feature: List[int]
    set_feature: Set[int]
    set_str_feature: Set[str]
    frozenset_feature: FrozenSet[str]
    homogenous_tuple_feature: Tuple[int, ...]
    list_of_struct_feature: List[MyNamedTuple]
    nested_struct: StructOfOptionalStructFeature


MISSING = object()


@dataclass
class Parameterization:
    wrapped_feature: Any
    expected_pa_dtype: pa.DataType
    expected_pl_dtype: pa.DataType
    expected_serialized_dtype: BaseSerializedDType

    rich: Any
    prim: Any
    json: Any

    canonical_rich: Any = MISSING

    should_error: bool = False

    @property
    def feature(self):
        return unwrap_feature(self.wrapped_feature)

    @property
    def pa(self):
        return pa.array([self.prim], self.expected_pa_dtype)

    def __post_init__(self):
        if self.canonical_rich is MISSING:
            self.canonical_rich = self.rich


@pytest.mark.parametrize(
    "p",
    [
        pytest.param(
            Parameterization(
                CodecFs.str_feature,
                pa.utf8(),
                pl.Utf8(),
                rich=None,
                prim=None,
                json=None,
                expected_serialized_dtype=SingletonDType(type_code=DTypeCode.STRING),
            ),
            id="str_null",
        ),
        pytest.param(
            Parameterization(
                CodecFs.str_feature,
                pa.utf8(),
                pl.Utf8(),
                rich="hi",
                prim="hi",
                json="hi",
                expected_serialized_dtype=SingletonDType(type_code=DTypeCode.STRING),
            ),
            id="str",
        ),
        pytest.param(
            Parameterization(
                CodecFs.bool_feature,
                pa.bool_(),
                pl.Boolean(),
                rich=True,
                prim=True,
                json=True,
                expected_serialized_dtype=SingletonDType(type_code=DTypeCode.BOOL),
            ),
            id="bool",
        ),
        pytest.param(
            Parameterization(
                CodecFs.bool_feature,
                pa.bool_(),
                pl.Boolean(),
                rich="t",
                canonical_rich=True,
                prim=True,
                json=True,
                expected_serialized_dtype=SingletonDType(type_code=DTypeCode.BOOL),
            ),
            id="bool_str",
        ),
        pytest.param(
            Parameterization(
                CodecFs.bool_feature,
                pa.bool_(),
                pl.Boolean(),
                rich=0,
                canonical_rich=False,
                prim=False,
                json=False,
                expected_serialized_dtype=SingletonDType(type_code=DTypeCode.BOOL),
            ),
            id="bool_int",
        ),
        pytest.param(
            Parameterization(
                CodecFs.int_feature,
                pa.int64(),
                pl.Int64(),
                rich=2,
                canonical_rich=2,
                prim=2,
                json=2,
                expected_serialized_dtype=SingletonDType(type_code=DTypeCode.INT64),
            ),
            id="int",
        ),
        pytest.param(
            Parameterization(
                CodecFs.int_feature,
                pa.int64(),
                pl.Int64(),
                rich=2.0,
                canonical_rich=2,
                prim=2,
                json=2,
                expected_serialized_dtype=SingletonDType(type_code=DTypeCode.INT64),
            ),
            id="float_as_int",
        ),
        pytest.param(
            Parameterization(
                CodecFs.unsigned_int_feature,
                pa.uint64(),
                pl.UInt64(),
                rich=2**64 - 1,
                prim=2**64 - 1,
                json=2**64 - 1,
                expected_serialized_dtype=SingletonDType(type_code=DTypeCode.UINT64),
            ),
            id="unsigned_int",
        ),
        pytest.param(
            Parameterization(
                CodecFs.float_feature,
                pa.float64(),
                pl.Float64(),
                rich=3.14,
                prim=3.14,
                json=3.14,
                expected_serialized_dtype=SingletonDType(type_code=DTypeCode.FLOAT64),
            ),
            id="float",
        ),
        pytest.param(
            Parameterization(
                CodecFs.float_feature,
                pa.float64(),
                pl.Float64(),
                canonical_rich=3.0,
                rich=3,
                prim=3.0,
                json=3.0,
                expected_serialized_dtype=SingletonDType(type_code=DTypeCode.FLOAT64),
            ),
            id="int_as_float",
        ),
        pytest.param(
            Parameterization(
                CodecFs.datetime_feature,
                pa.timestamp("us", "UTC"),
                pl.Datetime("us", "UTC"),
                rich=datetime(2000, 1, 1, 3, 45, 56, 789, timezone.utc),
                prim=datetime(2000, 1, 1, 3, 45, 56, 789, timezone.utc),
                json="2000-01-01T03:45:56.000789+00:00",
                expected_serialized_dtype=TimestampDType(
                    type_code=DTypeCode.TIMESTAMP,
                    time_unit=TimeUnitCode.MICROSECONDS,
                    timezone="UTC",
                ),
            ),
            id="datetime",
        ),
        pytest.param(
            Parameterization(
                CodecFs.datetime_feature,
                pa.timestamp("us", "UTC"),
                pl.Datetime("us", "UTC"),
                rich="2000-01-01T03:45:56.000789Z",
                canonical_rich=datetime(2000, 1, 1, 3, 45, 56, 789, timezone.utc),
                prim=datetime(2000, 1, 1, 3, 45, 56, 789, timezone.utc),
                json="2000-01-01T03:45:56.000789+00:00",
                expected_serialized_dtype=TimestampDType(
                    type_code=DTypeCode.TIMESTAMP,
                    time_unit=TimeUnitCode.MICROSECONDS,
                    timezone="UTC",
                ),
            ),
            id="str_as_datetime",
        ),
        pytest.param(
            Parameterization(
                CodecFs.datetime_without_timezone_feature,
                pa.timestamp("us"),
                pl.Datetime("us"),
                rich=datetime(2000, 1, 1, 3, 45, 56, 789),
                prim=datetime(2000, 1, 1, 3, 45, 56, 789),
                json="2000-01-01T03:45:56.000789",
                expected_serialized_dtype=TimestampDType(
                    type_code=DTypeCode.TIMESTAMP,
                    time_unit=TimeUnitCode.MICROSECONDS,
                    timezone=None,
                ),
            ),
            id="datetime_wo_timezone",
        ),
        pytest.param(
            Parameterization(
                CodecFs.datetime_without_timezone_feature,
                pa.timestamp("us"),
                pl.Datetime("us"),
                rich="2000-01-01T03:45:56.000789",
                canonical_rich=datetime(2000, 1, 1, 3, 45, 56, 789),
                prim=datetime(2000, 1, 1, 3, 45, 56, 789),
                json="2000-01-01T03:45:56.000789",
                expected_serialized_dtype=TimestampDType(
                    type_code=DTypeCode.TIMESTAMP,
                    time_unit=TimeUnitCode.MICROSECONDS,
                    timezone=None,
                ),
            ),
            id="str_datetime_wo_timezone",
        ),
        pytest.param(
            Parameterization(
                CodecFs.datetime_without_timezone_feature,
                pa.timestamp("us", None),
                pl.Datetime("us", None),
                rich="2023-01-25 17:31:40.074654",
                canonical_rich=datetime(2023, 1, 25, 17, 31, 40, 74654),
                prim=datetime(2023, 1, 25, 17, 31, 40, 74654),
                json="2023-01-25T17:31:40.074654",
                expected_serialized_dtype=TimestampDType(
                    type_code=DTypeCode.TIMESTAMP,
                    time_unit=TimeUnitCode.MICROSECONDS,
                    timezone=None,
                ),
            ),
            id="non_iso_str_as_datetime",
        ),
        pytest.param(
            Parameterization(
                CodecFs.datetime_without_timezone_feature,
                pa.timestamp("us", None),
                pl.Datetime("us", None),
                rich="2023-01-25",
                canonical_rich=datetime(2023, 1, 25),
                prim=datetime(2023, 1, 25),
                json="2023-01-25T00:00:00",
                expected_serialized_dtype=TimestampDType(
                    type_code=DTypeCode.TIMESTAMP,
                    time_unit=TimeUnitCode.MICROSECONDS,
                    timezone=None,
                ),
            ),
            id="date_str_as_datetime",
        ),
        pytest.param(
            Parameterization(
                CodecFs.time_feature,
                pa.time64("us"),
                pl.Time(),
                rich=time(3, 45, 56, 789),
                canonical_rich=time(3, 45, 56, 789),
                prim=time(3, 45, 56, 789),
                json="03:45:56.000789",
                expected_serialized_dtype=TimeUnitDType(
                    type_code=DTypeCode.TIME64,
                    time_unit=TimeUnitCode.MICROSECONDS,
                ),
            ),
            id="time",
        ),
        pytest.param(
            Parameterization(
                CodecFs.time_feature,
                pa.time64("us"),
                pl.Time(),
                rich=time(3, 45, 56, 789),
                canonical_rich=time(3, 45, 56, 789),
                prim=time(3, 45, 56, 789),
                json="03:45:56.000789",
                expected_serialized_dtype=TimeUnitDType(
                    type_code=DTypeCode.TIME64,
                    time_unit=TimeUnitCode.MICROSECONDS,
                ),
            ),
            id="str_time",
        ),
        pytest.param(
            Parameterization(
                CodecFs.date_feature,
                pa.date64(),
                # Polars parses dates as ms datetimes
                pl.Datetime("ms", None),
                rich=date(2000, 1, 1),
                prim=date(2000, 1, 1),
                json="2000-01-01",
                expected_serialized_dtype=SingletonDType(
                    type_code=DTypeCode.DATE64,
                ),
            ),
            id="datetime_date",
        ),
        pytest.param(
            Parameterization(
                CodecFs.date_feature,
                pa.date64(),
                # Polars parses dates as ms datetimes
                pl.Datetime("ms", None),
                rich="2000-01-01",
                canonical_rich=date(2000, 1, 1),
                prim=date(2000, 1, 1),
                json="2000-01-01",
                expected_serialized_dtype=SingletonDType(
                    type_code=DTypeCode.DATE64,
                ),
            ),
            id="str_date",
        ),
        pytest.param(
            Parameterization(
                CodecFs.date_feature,
                pa.date64(),
                # Polars parses dates as ms datetimes
                pl.Datetime("ms", None),
                rich=datetime(2000, 1, 1),
                canonical_rich=date(2000, 1, 1),
                prim=date(2000, 1, 1),
                json="2000-01-01",
                expected_serialized_dtype=SingletonDType(
                    type_code=DTypeCode.DATE64,
                ),
            ),
            id="date",
        ),
        pytest.param(
            Parameterization(
                CodecFs.duration_feature,
                pa.duration("us"),
                pl.Duration(),
                rich=timedelta(minutes=5),
                prim=timedelta(minutes=5),
                json="PT5M",
                expected_serialized_dtype=TimeUnitDType(
                    type_code=DTypeCode.DURATION,
                    time_unit=TimeUnitCode.MICROSECONDS,
                ),
            ),
            id="timedelta",
        ),
        pytest.param(
            Parameterization(
                CodecFs.duration_feature,
                pa.duration("us"),
                pl.Duration(),
                rich="P0Y0MT5M0S",
                canonical_rich=timedelta(minutes=5),
                prim=timedelta(minutes=5),
                json="PT5M",
                expected_serialized_dtype=TimeUnitDType(
                    type_code=DTypeCode.DURATION,
                    time_unit=TimeUnitCode.MICROSECONDS,
                ),
            ),
            id="timedelta_iso",
        ),
        pytest.param(
            Parameterization(
                CodecFs.binary_feature,
                pa.binary(),
                pl.Binary(),
                rich=b"Hello",
                prim=b"Hello",
                json=base64.b64encode(b"Hello").decode("utf8"),
                expected_serialized_dtype=BinaryDType(
                    type_code=DTypeCode.BINARY,
                    length=-1,
                ),
            ),
            id="binary",
        ),
        pytest.param(
            Parameterization(
                # We encode decimals as strings, as polars does not support
                # pyarrow decimal
                CodecFs.decimal_feature,
                pa.utf8(),
                pl.Utf8(),
                rich=Decimal("0.33"),
                prim="0.33",
                json="0.33",
                expected_serialized_dtype=SingletonDType(
                    type_code=DTypeCode.STRING,
                ),
            ),
            id="decimal",
        ),
        pytest.param(
            Parameterization(
                CodecFs.decimal_feature,
                pa.utf8(),
                pl.Utf8(),
                rich=Decimal("2.0000"),
                canonical_rich=Decimal("2"),
                prim="2",
                json="2",
                expected_serialized_dtype=SingletonDType(
                    type_code=DTypeCode.STRING,
                ),
            ),
            id="decimal_no_sig_figs",
        ),
        pytest.param(
            Parameterization(
                CodecFs.decimal_feature,
                pa.utf8(),
                pl.Utf8(),
                rich=2,
                canonical_rich=Decimal("2"),
                prim="2",
                json="2",
                expected_serialized_dtype=SingletonDType(
                    type_code=DTypeCode.STRING,
                ),
            ),
            id="decimal_int",
        ),
        pytest.param(
            Parameterization(
                CodecFs.decimal_feature,
                pa.utf8(),
                pl.Utf8(),
                rich=2.0,
                canonical_rich=Decimal("2"),
                prim="2",
                json="2",
                expected_serialized_dtype=SingletonDType(
                    type_code=DTypeCode.STRING,
                ),
            ),
            id="decimal_float",
        ),
        pytest.param(
            Parameterization(
                CodecFs.decimal_feature,
                pa.utf8(),
                pl.Utf8(),
                rich="2.0",
                canonical_rich=Decimal("2"),
                prim="2",
                json="2",
                expected_serialized_dtype=SingletonDType(
                    type_code=DTypeCode.STRING,
                ),
            ),
            id="decimal_str",
        ),
        pytest.param(
            Parameterization(
                CodecFs.str_enum_feature,
                pa.utf8(),
                pl.Utf8(),
                rich=StrEnumFeature.BLUE,
                prim="blue",
                json="blue",
                expected_serialized_dtype=SingletonDType(
                    type_code=DTypeCode.STRING,
                ),
            ),
            id="str_enum",
        ),
        pytest.param(
            Parameterization(
                CodecFs.str_enum_feature,
                pa.utf8(),
                pl.Utf8(),
                rich="blue",
                canonical_rich=StrEnumFeature.BLUE,
                prim="blue",
                json="blue",
                expected_serialized_dtype=SingletonDType(
                    type_code=DTypeCode.STRING,
                ),
            ),
            id="str_enum_val",
        ),
        pytest.param(
            Parameterization(
                CodecFs.str_enum_feature,
                pa.utf8(),
                pl.Utf8(),
                rich="BLUE",
                canonical_rich=StrEnumFeature.BLUE,
                prim="blue",
                json="blue",
                expected_serialized_dtype=SingletonDType(
                    type_code=DTypeCode.STRING,
                ),
            ),
            id="str_enum_key",
        ),
        pytest.param(
            Parameterization(
                CodecFs.int_enum_feature,
                pa.int64(),
                pl.Int64(),
                rich=IntEnumFeature.ONE,
                prim=1,
                json=1,
                expected_serialized_dtype=SingletonDType(
                    type_code=DTypeCode.INT64,
                ),
            ),
            id="int_enum",
        ),
        pytest.param(
            Parameterization(
                CodecFs.int_enum_feature,
                pa.int64(),
                pl.Int64(),
                rich=1,
                canonical_rich=IntEnumFeature.ONE,
                prim=1,
                json=1,
                expected_serialized_dtype=SingletonDType(
                    type_code=DTypeCode.INT64,
                ),
            ),
            id="int_enum_val",
        ),
        pytest.param(
            Parameterization(
                CodecFs.int_enum_feature,
                pa.int64(),
                pl.Int64(),
                rich="ONE",
                canonical_rich=IntEnumFeature.ONE,
                prim=1,
                json=1,
                expected_serialized_dtype=SingletonDType(
                    type_code=DTypeCode.INT64,
                ),
            ),
            id="int_enum_key",
        ),
        pytest.param(
            Parameterization(
                CodecFs.dataclass_feature,
                PA_STRUCT_SCHEMA,
                PL_STRUCT_SCHEMA,
                rich=None,
                canonical_rich=MyDataclass(
                    foo=cast(int, None), bar=cast(str, None), baz=None, sub_list=cast(list, None)
                ),
                prim={"foo": None, "bar": None, "baz": None, "sub_list": None},
                json=[None, None, None, None],
                expected_serialized_dtype=SERIALIZED_STRUCT_DTYPE,
            ),
            id="null_struct",
        ),
        pytest.param(
            Parameterization(
                CodecFs.dataclass_feature,
                PA_STRUCT_SCHEMA,
                PL_STRUCT_SCHEMA,
                rich=(None, None, None, None),
                canonical_rich=MyDataclass(
                    foo=cast(int, None), bar=cast(str, None), baz=None, sub_list=cast(list, None)
                ),
                prim={"foo": None, "bar": None, "baz": None, "sub_list": None},
                json=[None, None, None, None],
                expected_serialized_dtype=SERIALIZED_STRUCT_DTYPE,
            ),
            id="dataclass_struct_null_tuple",
        ),
        pytest.param(
            Parameterization(
                CodecFs.dataclass_feature,
                PA_STRUCT_SCHEMA,
                PL_STRUCT_SCHEMA,
                rich=MyDataclass(foo=cast(int, None), bar=cast(str, None), baz=None, sub_list=cast(list, None)),
                prim={"foo": None, "bar": None, "baz": None, "sub_list": None},
                json=[None, None, None, None],
                expected_serialized_dtype=SERIALIZED_STRUCT_DTYPE,
            ),
            id="dataclass_struct_null",
        ),
        pytest.param(
            Parameterization(
                CodecFs.dataclass_feature,
                PA_STRUCT_SCHEMA,
                PL_STRUCT_SCHEMA,
                rich=MyDataclass(50, "hi", None, [100]),
                prim={"foo": 50, "bar": "hi", "baz": None, "sub_list": [100]},
                json=[50, "hi", None, [100]],
                expected_serialized_dtype=SERIALIZED_STRUCT_DTYPE,
            ),
            id="dataclass_struct",
        ),
        pytest.param(
            Parameterization(
                CodecFs.named_tuple_feature,
                PA_STRUCT_SCHEMA,
                PL_STRUCT_SCHEMA,
                rich=MyNamedTuple(50, "hi", None, [100]),
                prim={"foo": 50, "bar": "hi", "baz": None, "sub_list": [100]},
                json=[50, "hi", None, [100]],
                expected_serialized_dtype=SERIALIZED_STRUCT_DTYPE,
            ),
            id="namedtuple_struct",
        ),
        pytest.param(
            Parameterization(
                CodecFs.named_tuple_feature,
                PA_STRUCT_SCHEMA,
                PL_STRUCT_SCHEMA,
                rich=MyNamedTuple(foo=cast(int, None), bar=cast(str, None), baz=None, sub_list=cast(list, None)),
                prim={"foo": None, "bar": None, "baz": None, "sub_list": None},
                json=[None, None, None, None],
                expected_serialized_dtype=SERIALIZED_STRUCT_DTYPE,
            ),
            id="namedtuple_struct_nulls",
        ),
        pytest.param(
            Parameterization(
                CodecFs.pydantic_feature,
                PA_STRUCT_SCHEMA,
                PL_STRUCT_SCHEMA,
                rich=PydanticFeature(foo=50, bar="hi", baz=None, sub_list=[100]),
                prim={"foo": 50, "bar": "hi", "baz": None, "sub_list": [100]},
                json=[50, "hi", None, [100]],
                expected_serialized_dtype=SERIALIZED_STRUCT_DTYPE,
            ),
            id="pydantic_struct",
        ),
        pytest.param(
            Parameterization(
                CodecFs.pydantic_feature,
                PA_STRUCT_SCHEMA,
                PL_STRUCT_SCHEMA,
                rich=PydanticFeature.construct(foo=None, bar=None, baz=None, sub_list=None),
                prim={"foo": None, "bar": None, "baz": None, "sub_list": None},
                json=[None, None, None, None],
                expected_serialized_dtype=SERIALIZED_STRUCT_DTYPE,
            ),
            id="pydantic_struct_nulls",
        ),
        pytest.param(
            Parameterization(
                CodecFs.typed_dict_feature,
                PA_STRUCT_SCHEMA,
                PL_STRUCT_SCHEMA,
                rich={"foo": 50, "bar": "hi", "baz": None, "sub_list": [100]},
                prim={"foo": 50, "bar": "hi", "baz": None, "sub_list": [100]},
                json=[50, "hi", None, [100]],
                expected_serialized_dtype=SERIALIZED_STRUCT_DTYPE,
            ),
            id="typed_dict_struct",
        ),
        pytest.param(
            Parameterization(
                CodecFs.typed_dict_feature,
                PA_STRUCT_SCHEMA,
                PL_STRUCT_SCHEMA,
                rich={"foo": None, "bar": None, "baz": None, "sub_list": None},
                prim={"foo": None, "bar": None, "baz": None, "sub_list": None},
                json=[None, None, None, None],
                expected_serialized_dtype=SERIALIZED_STRUCT_DTYPE,
            ),
            id="typed_dict_struct_nulls",
        ),
        pytest.param(
            Parameterization(
                CodecFs.attrs_feature,
                PA_STRUCT_SCHEMA,
                PL_STRUCT_SCHEMA,
                rich=AttrsFeature(foo=50, bar="hi", baz=None, sub_list=[100]),
                prim={"foo": 50, "bar": "hi", "baz": None, "sub_list": [100]},
                json=[50, "hi", None, [100]],
                expected_serialized_dtype=SERIALIZED_STRUCT_DTYPE,
            ),
            id="attrs_struct",
        ),
        pytest.param(
            Parameterization(
                CodecFs.attrs_feature,
                PA_STRUCT_SCHEMA,
                PL_STRUCT_SCHEMA,
                rich=AttrsFeature(foo=cast(int, None), bar=cast(str, None), baz=None, sub_list=cast(list, None)),
                prim={"foo": None, "bar": None, "baz": None, "sub_list": None},
                json=[None, None, None, None],
                expected_serialized_dtype=SERIALIZED_STRUCT_DTYPE,
            ),
            id="attrs_struct_nulls",
        ),
        pytest.param(
            Parameterization(
                CodecFs.list_feature,
                pa.list_(pa.int64()),
                pl.List(pl.Int64()),
                rich=None,
                prim=None,
                json=None,
                expected_serialized_dtype=ListDType(
                    type_code=DTypeCode.LIST,
                    inner_dtype=SingletonDType(type_code=DTypeCode.INT64),
                    length=-1,
                ),
            ),
            id="null_list",
        ),
        pytest.param(
            Parameterization(
                CodecFs.list_feature,
                pa.list_(pa.int64()),
                pl.List(pl.Int64()),
                rich=[None],
                prim=[None],
                json=[None],
                expected_serialized_dtype=ListDType(
                    type_code=DTypeCode.LIST,
                    inner_dtype=SingletonDType(type_code=DTypeCode.INT64),
                    length=-1,
                ),
            ),
            id="list_of_null",
        ),
        pytest.param(
            Parameterization(
                CodecFs.list_feature,
                pa.list_(pa.int64()),
                pl.List(pl.Int64()),
                rich=[2, 3],
                prim=[2, 3],
                json=[2, 3],
                expected_serialized_dtype=ListDType(
                    type_code=DTypeCode.LIST,
                    inner_dtype=SingletonDType(type_code=DTypeCode.INT64),
                    length=-1,
                ),
            ),
            id="real_list",
        ),
        pytest.param(
            Parameterization(
                CodecFs.set_feature,
                pa.list_(pa.int64()),
                pl.List(pl.Int64()),
                rich=set((2, 3)),
                prim=[2, 3],
                json=[2, 3],
                expected_serialized_dtype=ListDType(
                    type_code=DTypeCode.LIST,
                    inner_dtype=SingletonDType(type_code=DTypeCode.INT64),
                    length=-1,
                ),
            ),
            id="set",
        ),
        pytest.param(
            Parameterization(
                CodecFs.set_feature,
                pa.list_(pa.int64()),
                pl.List(pl.Int64()),
                # even though the insertion order is (3, 2), it should be serialized in the same order
                # as the above test, which is (2, 3)
                rich=set((3, 2)),
                prim=[2, 3],
                json=[2, 3],
                expected_serialized_dtype=ListDType(
                    type_code=DTypeCode.LIST,
                    inner_dtype=SingletonDType(type_code=DTypeCode.INT64),
                    length=-1,
                ),
            ),
            id="set_ordering",
        ),
        pytest.param(
            Parameterization(
                CodecFs.set_str_feature,
                pa.list_(pa.utf8()),
                pl.List(pl.Utf8()),
                rich=set(("c", "a", "b")),
                prim=["a", "b", "c"],
                json=["a", "b", "c"],
                expected_serialized_dtype=ListDType(
                    type_code=DTypeCode.LIST,
                    inner_dtype=SingletonDType(type_code=DTypeCode.STRING),
                    length=-1,
                ),
            ),
            id="set_ordering_str",
        ),
        pytest.param(
            Parameterization(
                CodecFs.frozenset_feature,
                pa.list_(pa.utf8()),
                pl.List(pl.Utf8()),
                rich=frozenset(("c", "a", "b")),
                prim=["a", "b", "c"],
                json=["a", "b", "c"],
                expected_serialized_dtype=ListDType(
                    type_code=DTypeCode.LIST,
                    inner_dtype=SingletonDType(type_code=DTypeCode.STRING),
                    length=-1,
                ),
            ),
            id="frozenset_feature",
        ),
        pytest.param(
            Parameterization(
                CodecFs.homogenous_tuple_feature,
                pa.list_(pa.int64()),
                pl.List(pl.Int64()),
                rich=(2, 3),
                prim=[2, 3],
                expected_serialized_dtype=ListDType(
                    type_code=DTypeCode.LIST,
                    inner_dtype=SingletonDType(type_code=DTypeCode.INT64),
                    length=-1,
                ),
                json=[2, 3],
            ),
            id="homogenous_tuple_feature",
        ),
        pytest.param(
            Parameterization(
                CodecFs.custom_feature,
                PA_STRUCT_SCHEMA,
                PL_STRUCT_SCHEMA,
                rich=CustomClass(foo=50, bar="hi", baz=None, sub_list=[100]),
                prim={"foo": 50, "bar": "hi", "baz": None, "sub_list": [100]},
                json=[50, "hi", None, [100]],
                expected_serialized_dtype=SERIALIZED_STRUCT_DTYPE,
            ),
            id="custom",
        ),
        pytest.param(
            Parameterization(
                CodecFs.custom_feature,
                PA_STRUCT_SCHEMA,
                PL_STRUCT_SCHEMA,
                rich=CustomClass(foo=cast(int, None), bar=cast(str, None), baz=None, sub_list=cast(list, None)),
                prim={"foo": None, "bar": None, "baz": None, "sub_list": None},
                json=[None, None, None, None],
                expected_serialized_dtype=SERIALIZED_STRUCT_DTYPE,
            ),
            id="custom_nulls",
        ),
        pytest.param(
            Parameterization(
                CodecFs.custom_feature,
                PA_STRUCT_SCHEMA,
                PL_STRUCT_SCHEMA,
                rich=None,
                canonical_rich=CustomClass(
                    foo=cast(int, None), bar=cast(str, None), baz=None, sub_list=cast(list, None)
                ),
                prim={"foo": None, "bar": None, "baz": None, "sub_list": None},
                json=[None, None, None, None],
                expected_serialized_dtype=SERIALIZED_STRUCT_DTYPE,
            ),
            id="null_custom",
        ),
        pytest.param(
            Parameterization(
                CodecFs.list_of_struct_feature,
                pa.list_(PA_STRUCT_SCHEMA),
                pl.List(PL_STRUCT_SCHEMA),
                rich=None,
                prim=None,
                json=None,
                expected_serialized_dtype=ListDType(
                    type_code=DTypeCode.LIST,
                    inner_dtype=SERIALIZED_STRUCT_DTYPE,
                    length=-1,
                ),
            ),
            id="null_list_of_struct",
        ),
        pytest.param(
            Parameterization(
                CodecFs.list_of_struct_feature,
                pa.list_(PA_STRUCT_SCHEMA),
                pl.List(PL_STRUCT_SCHEMA),
                rich=[],
                prim=[],
                json=[],
                expected_serialized_dtype=ListDType(
                    type_code=DTypeCode.LIST,
                    inner_dtype=SERIALIZED_STRUCT_DTYPE,
                    length=-1,
                ),
            ),
            id="list_of_struct_empty",
        ),
        pytest.param(
            Parameterization(
                CodecFs.list_of_struct_feature,
                pa.list_(PA_STRUCT_SCHEMA),
                pl.List(PL_STRUCT_SCHEMA),
                rich=[MyNamedTuple(foo=cast(int, None), bar=cast(str, None), baz=None, sub_list=cast(list, None))],
                prim=[{"foo": None, "bar": None, "baz": None, "sub_list": None}],
                json=[[None, None, None, None]],
                expected_serialized_dtype=ListDType(
                    type_code=DTypeCode.LIST,
                    inner_dtype=SERIALIZED_STRUCT_DTYPE,
                    length=-1,
                ),
            ),
            id="list_of_null_struct",
        ),
        pytest.param(
            Parameterization(
                CodecFs.list_of_struct_feature,
                pa.list_(PA_STRUCT_SCHEMA),
                pl.List(PL_STRUCT_SCHEMA),
                rich=[MyNamedTuple(50, "hi", None, [100])],
                prim=[{"foo": 50, "bar": "hi", "baz": None, "sub_list": [100]}],
                json=[[50, "hi", None, [100]]],
                expected_serialized_dtype=ListDType(
                    type_code=DTypeCode.LIST,
                    inner_dtype=SERIALIZED_STRUCT_DTYPE,
                    length=-1,
                ),
            ),
            id="list_of_struct",
        ),
        pytest.param(
            Parameterization(
                CodecFs.nested_struct,
                pa.struct([pa.field("sub", pa.struct([pa.field("name", pa.utf8())]))]),
                pl.Struct([pl.Field("sub", pl.Struct([pl.Field("name", pl.Utf8())]))]),
                rich={"sub": None},
                canonical_rich={"sub": {"name": None}},
                prim={"sub": {"name": None}},
                json=[[None]],
                expected_serialized_dtype=StructDType(
                    type_code=DTypeCode.STRUCT,
                    fields=[
                        StructField(
                            name="sub",
                            dtype=StructDType(
                                type_code=DTypeCode.STRUCT,
                                fields=[
                                    StructField(
                                        name="name",
                                        dtype=SingletonDType(type_code=DTypeCode.STRING),
                                    ),
                                ],
                            ),
                        )
                    ],
                ),
            ),
            id="sub_struct",
        ),
    ],
)
class TestEncoder:
    def test_infer_pa_dtype(self, p: Parameterization):
        assert p.feature.converter.pyarrow_dtype == p.expected_pa_dtype

    def test_infer_pl_dtype(self, p: Parameterization):
        assert p.feature.converter.polars_dtype == p.expected_pl_dtype

    def test_rich_to_pyarrow(self, p: Parameterization):
        assert p.feature.converter.from_rich_to_pyarrow([p.rich]) == p.pa

    def test_rich_to_prim(self, p: Parameterization):
        assert p.feature.converter.from_rich_to_primitive(p.rich) == p.prim

    def test_rich_to_json(self, p: Parameterization):
        assert p.feature.converter.from_rich_to_json(p.rich) == p.json

    def test_prim_to_pyarrow(self, p: Parameterization):
        assert p.feature.converter.from_primitive_to_pyarrow([p.prim]) == p.pa

    def test_prim_to_rich(self, p: Parameterization):
        converted = p.feature.converter.from_primitive_to_rich(p.prim)
        assert isinstance(converted, type(p.canonical_rich))
        assert converted == p.canonical_rich

    def test_prim_to_json(self, p: Parameterization):
        assert p.feature.converter.from_primitive_to_json(p.prim) == p.json

    def test_pyarrow_to_rich(self, p: Parameterization):
        converted = p.feature.converter.from_pyarrow_to_rich(p.pa)
        assert len(converted) == 1
        x = converted[0]
        assert isinstance(x, type(p.canonical_rich))
        assert x == p.canonical_rich

    def test_pyarrow_to_prim(self, p: Parameterization):
        assert p.feature.converter.from_pyarrow_to_primitive(p.pa) == [p.prim]

    def test_pyarrow_to_json(self, p: Parameterization):
        assert p.feature.converter.from_pyarrow_to_json(p.pa) == [p.json]

    def test_json_to_pyarrow(self, p: Parameterization):
        assert p.feature.converter.from_json_to_pyarrow([p.json]) == p.pa

    def test_json_to_prim(self, p: Parameterization):
        assert p.feature.converter.from_json_to_primitive(p.json) == p.prim

    def test_json_to_rich(self, p: Parameterization):
        converted = p.feature.converter.from_json_to_rich(p.json)
        assert isinstance(converted, type(p.canonical_rich))
        assert converted == p.canonical_rich

    def test_rich_to_dataframe(self, p: Parameterization):
        df = DataFrame(
            {
                p.feature: [p.rich],
            }
        )
        self._validate_df(df, p)

    def _validate_df(self, df: DataFrame, p: Parameterization):
        fs = df.to_features()
        assert len(fs) == 1
        val = getattr(fs[0], p.feature.attribute_name)
        assert isinstance(val, type(p.canonical_rich))
        assert val == p.canonical_rich

    def test_prim_to_dataframe(self, p: Parameterization):
        df = DataFrame(
            {
                p.feature: [p.prim],
            }
        )
        self._validate_df(df, p)

    def test_df_from_pyarrow(self, p: Parameterization):
        series = pl.from_arrow(p.pa)
        assert isinstance(series, pl.Series)
        pl_df = pl.DataFrame({p.feature.fqn: series})
        df = DataFrame(pl_df)
        self._validate_df(df, p)

    def test_serialized_dtype(self, p: Parameterization):
        assert p.feature.converter.serialized_dtype == p.expected_serialized_dtype
        assert p.feature.converter.serialized_dtype.to_pyarrow_dtype() == p.expected_pa_dtype



import json
from pathlib import Path

import pytest

from chalk.features import feature, features, has_one
from chalk.parsed.user_types_to_json import get_registered_types_as_json


@features
class VChild1:
    id: str
    evolving: int = feature(version=2, default_version=1)


@features
class VChild:
    id: str
    evolving: int = feature(version=2, default_version=2)


@features
class VParent:
    id: str
    child: VChild = has_one(lambda: VChild.id == VParent.id)
    child1: VChild1 = has_one(lambda: VChild1.id == VParent.id)


@features
class VGrandParent:
    id: str
    parent: VParent = has_one(lambda: VGrandParent.id == VParent.id)


def test_versioned_feature_names():
    assert str(VChild.evolving) == "v_child.evolving:v2"
    assert str(VChild.evolving_v2) == "v_child.evolving:v2"
    assert str(VChild.evolving_v1) == "v_child.evolving"
    assert str(VParent.child.evolving) == "v_parent.child.evolving:v2"
    assert str(VGrandParent.parent.child.evolving) == "v_grand_parent.parent.child.evolving:v2"


def test_applying_versions():
    assert str(VParent.child.evolving @ 2) == "v_parent.child.evolving:v2"
    assert str(VGrandParent.parent.child.evolving @ 2) == "v_grand_parent.parent.child.evolving:v2"

    assert str(VParent.child.evolving @ 1) == "v_parent.child.evolving"
    assert str(VGrandParent.parent.child.evolving @ 1) == "v_grand_parent.parent.child.evolving"


def test_applying_versions_default_1():
    assert str(VParent.child1.evolving @ 2) == "v_parent.child1.evolving:v2"
    assert str(VGrandParent.parent.child1.evolving @ 2) == "v_grand_parent.parent.child1.evolving:v2"

    assert str(VParent.child1.evolving @ 1) == "v_parent.child1.evolving"
    assert str(VGrandParent.parent.child1.evolving @ 1) == "v_grand_parent.parent.child1.evolving"


def test_unversioned():
    with pytest.raises(ValueError) as e:
        str(VParent.id @ 3)

    assert str(e.value) == (
        "Cannot request version 3 of feature 'v_parent.id', because this feature doesn't "
        "have a version set at definition. To set a version, write \n"
        "    @features\n"
        "    class VParent:\n"
        "        id: ... = feature(version=3)\n"
        "        ...\n"
    )


def test_too_many_versions():
    with pytest.raises(ValueError) as e:
        str(VParent.child.evolving @ 3)

    assert str(e.value) == (
        "Cannot request version 3 of feature 'v_parent.child.evolving', because "
        "this feature has a maximum version of 2 < 3. To add versions, write \n"
        "    @features\n"
        "    class VChild:\n"
        "   -    evolving: ... = feature(version=2)\n"
        "   +    evolving: ... = feature(version=3)\n"
        "        ...\n"
    )


def test_supplying_two_versions():
    dual_versioned = VChild(evolving_v2=2, evolving_v1=1)
    assert dual_versioned.evolving_v2 == 2
    assert dual_versioned.evolving_v1 == 1
    # The default is v2
    assert dual_versioned.evolving == 2

    dual_versioned = VChild1(evolving_v2=2, evolving_v1=1)
    assert dual_versioned.evolving_v2 == 2
    assert dual_versioned.evolving_v1 == 1
    # The default is v2
    assert dual_versioned.evolving == 1


def test_aliases_unique():
    with pytest.raises(ValueError) as e:
        VChild(evolving=2, evolving_v2=1)

    assert (
        str(e.value)
        == "The features 'evolving' and 'evolving_v2' are aliases of each other. Only one can be specified, but both were given."
    )


def test_default_constructor():
    assert VChild(evolving=2, evolving_v1=1).evolving == 2
    assert VChild1(evolving=2, evolving_v2=1).evolving == 2
    assert VChild(evolving_v2=2, evolving_v1=1).evolving == 2


def test_setters():
    existing = VChild(evolving=2)
    assert existing.evolving == 2
    existing.evolving_v2 = 5
    assert existing.evolving_v2 == 5
    assert existing.evolving == 5

    existing = VChild(evolving=2)
    existing.evolving = 5
    assert existing.evolving == 5
    assert existing.evolving_v2 == 5


def test_collision():
    with pytest.raises(ValueError) as e:

        @features
        class X:
            v: int = feature(version=2)
            v_v2: int

    assert e.value.args[0] == (
        "The class 'X' has an existing annotation 'v_v2' "
        "that collides with a versioned feature. Please remove "
        "the existing annotation, or lower the version."
    )


@features
class Funky:
    id: str
    trumpet: str = feature(version=2)
    saxophone: str = feature(version=3, default_version=3)


def test_version_iter():
    assert dict(Funky(trumpet="red", saxophone_v1="purple")) == {
        "funky.trumpet": "red",
        "funky.saxophone": "purple",
    }

    assert dict(Funky(trumpet="red", saxophone="purple")) == {
        "funky.trumpet": "red",
        "funky.saxophone:v3": "purple",
    }


def test_len():
    assert len(Funky(trumpet="red", saxophone="purple")) == 2


@pytest.mark.filterwarnings("ignore:`feature.typ.underlying`:DeprecationWarning")
def test_registry():
    fqns = [
        feature_json["id"]["fqn"]
        for feature_json in json.loads(
            get_registered_types_as_json(
                Path("/"),
                failed=[],
            )
        )["features"]
        if feature_json["id"]["namespace"] == "funky"
    ]
    assert sorted(fqns) == [
        "funky.id",
        "funky.saxophone",
        "funky.saxophone:v2",
        "funky.saxophone:v3",
        "funky.trumpet",
        "funky.trumpet:v2",
    ]



try:
    import polars as pl
except ImportError:
    pl = None

from datetime import datetime

import pytest

from chalk.client.dataset import DatasetVersion, _extract_df_columns
from chalk.features import feature, feature_time, features, has_one


@features
class TransactionV2:
    id: int = feature(primary=True)
    user_id: int
    amount: float
    ts: datetime = feature_time()


@features
class TransactionPointer:
    id: int = feature(primary=True)
    transaction_id: int = feature(max_staleness="infinity")
    transaction: TransactionV2 = has_one(lambda: TransactionV2.id == TransactionPointer.transaction_id)


@pytest.mark.skipif(pl is None, reason="Polars is not installed")
def test_feature_explosion_load_dataset():
    schema = [
        "__id__",
        "__oat__",
        "__rat__",
        "__ts__",
        "cb_orzgc3ttmfrxi2lpnyxgc3lpovxhi000",
        "cb_orzgc3ttmfrxi2lpnyxgc3lpovxhils7l5xwc5c7l4000000",
        "cb_orzgc3ttmfrxi2lpnyxgc3lpovxhils7l5zgc5c7l4000000",
        "cb_orzgc3ttmfrxi2lpnyxgsza0",
        "cb_orzgc3ttmfrxi2lpnyxgszbol5pw6ylul5pq0000",
        "cb_orzgc3ttmfrxi2lpnyxgszbol5pxeylul5pq0000",
        "cb_orzgc3ttmfrxi2lpnyxhi4y0",
        "cb_orzgc3ttmfrxi2lpnyxhk43fojpwsza0",
        "cb_orzgc3ttmfrxi2lpnyxhk43fojpwszbol5pw6ylul5pq0000",
        "cb_orzgc3ttmfrxi2lpnyxhk43fojpwszbol5pxeylul5pq0000",
    ]
    data = [None] * len(schema)
    df = pl.DataFrame([data], columns=schema)
    result = _extract_df_columns(
        df=df,
        output_feature_fqns=["transaction"],
        output_ts=False,
        output_id=False,
        version=DatasetVersion.BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES,
    )
    assert result.columns == ["transaction.amount", "transaction.id", "transaction.ts", "transaction.user_id"]

    schema = [
        "__id__",
        "__oat__",
        "__rat__",
        "__ts__",
        "cb_orzgc3ttmfrxi2lpnzpxa33jnz2gk4rol5pwg2dbnrvv633consxe5tfmrpwc5c7l4000000",
        "cb_orzgc3ttmfrxi2lpnzpxa33jnz2gk4ronfsa0000",
        "cb_orzgc3ttmfrxi2lpnzpxa33jnz2gk4ronfsc4x27n5qxix27",
        "cb_orzgc3ttmfrxi2lpnzpxa33jnz2gk4ronfsc4x27ojqxix27",
        "cb_orzgc3ttmfrxi2lpnzpxa33jnz2gk4roorzgc3ttmfrxi2lpnyxgc3lpovxhi000",
        "cb_orzgc3ttmfrxi2lpnzpxa33jnz2gk4roorzgc3ttmfrxi2lpnyxgc3lpovxhils7l5xwc5c7l4000000",
        "cb_orzgc3ttmfrxi2lpnzpxa33jnz2gk4roorzgc3ttmfrxi2lpnyxgc3lpovxhils7l5zgc5c7l4000000",
        "cb_orzgc3ttmfrxi2lpnzpxa33jnz2gk4roorzgc3ttmfrxi2lpnyxgsza0",
        "cb_orzgc3ttmfrxi2lpnzpxa33jnz2gk4roorzgc3ttmfrxi2lpnyxgszbol5pw6ylul5pq0000",
        "cb_orzgc3ttmfrxi2lpnzpxa33jnz2gk4roorzgc3ttmfrxi2lpnyxgszbol5pxeylul5pq0000",
        "cb_orzgc3ttmfrxi2lpnzpxa33jnz2gk4roorzgc3ttmfrxi2lpnyxhk43fojpwsza0",
        "cb_orzgc3ttmfrxi2lpnzpxa33jnz2gk4roorzgc3ttmfrxi2lpnyxhk43fojpwszbol5pw6ylul5pq0000",
        "cb_orzgc3ttmfrxi2lpnzpxa33jnz2gk4roorzgc3ttmfrxi2lpnyxhk43fojpwszbol5pxeylul5pq0000",
        "cb_orzgc3ttmfrxi2lpnzpxa33jnz2gk4roorzgc3ttmfrxi2lpnzpwsza0",
        "cb_orzgc3ttmfrxi2lpnzpxa33jnz2gk4roorzgc3ttmfrxi2lpnzpwszbol5pw6ylul5pq0000",
        "cb_orzgc3ttmfrxi2lpnzpxa33jnz2gk4roorzgc3ttmfrxi2lpnzpwszbol5pxeylul5pq0000",
    ]
    data = [None] * len(schema)
    df = pl.DataFrame([data], columns=schema)
    result = _extract_df_columns(
        df=df,
        output_feature_fqns=["transaction_pointer.transaction"],
        output_ts=False,
        output_id=False,
        version=DatasetVersion.BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES,
    )
    assert result.columns == [
        "transaction_pointer.transaction.amount",
        "transaction_pointer.transaction.id",
        "transaction_pointer.transaction.user_id",
    ]



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Generic, Optional, Protocol, Sequence, Type, TypeVar, Union, cast

import pyarrow as pa

from chalk.features._encoding.json import structure_json_to_primitive, unstructure_primitive_to_json
from chalk.features._encoding.primitive import TPrimitive
from chalk.features._encoding.pyarrow import pyarrow_to_polars, pyarrow_to_primitive, rich_to_pyarrow
from chalk.features._encoding.rich import structure_primitive_to_rich, unstructure_rich_to_primitive
from chalk.features._encoding.serialized_dtype import serialize_pyarrow_dtype
from chalk.utils.json import TJSON

if TYPE_CHECKING:
    import polars as pl

_TRich = TypeVar("_TRich")
_TRichCo = TypeVar("_TRichCo", covariant=True)
_TRichCon = TypeVar("_TRichCon", contravariant=True)

_TPrim = TypeVar("_TPrim", bound=TPrimitive)
_TPrimCo = TypeVar("_TPrimCo", bound=TPrimitive, covariant=True)
_TPrimCon = TypeVar("_TPrimCon", bound=TPrimitive, contravariant=True)


class TEncoder(Protocol[_TPrimCo, _TRichCon]):
    def __call__(self, value: _TRichCon, /) -> _TPrimCo:
        ...


class TDecoder(Protocol[_TPrimCon, _TRichCo]):
    def __call__(self, value: _TPrimCon, /) -> _TRichCo:
        ...


class _UnknownType:
    ...


_UNKNOWN = _UnknownType()


class FeatureConverter(Generic[_TPrim, _TRich]):
    def __init__(
        self,
        name: str,
        rich_type: Union[Type[_TRich], _UnknownType] = _UNKNOWN,
        pyarrow_dtype: Optional[pa.DataType] = None,
        encoder: Optional[TEncoder[_TPrim, _TRich]] = None,
        decoder: Optional[TDecoder[_TPrim, _TRich]] = None,
    ) -> None:
        self._name = name
        self._rich_type = rich_type
        if pyarrow_dtype is None:
            if isinstance(rich_type, _UnknownType):
                raise ValueError("Either the `rich_type` or `pyarrow_dtype` must be provided")

            pyarrow_dtype = rich_to_pyarrow(rich_type, name)
        self._pyarrow_dtype = pyarrow_dtype
        self._primitive_type = pyarrow_to_primitive(self._pyarrow_dtype, name)
        if isinstance(rich_type, _UnknownType):
            if encoder is not None:
                raise ValueError("An encoder cannot be specified without also specifying the `rich_type`")
            if decoder is not None:
                raise ValueError("An encoder cannot be specified without also specifying the `rich_type`")
        self._encoder = encoder
        self._decoder = decoder
        self._serialized_dtype = serialize_pyarrow_dtype(self._pyarrow_dtype)

    def _to_primitive(self, val: Optional[_TRich]) -> Optional[_TPrim]:
        if val is None or self._encoder is None:
            # Structuring null values to the primitive type to ensure that a singular null for an entire struct
            # is propagated to individual struct fields -- e.g.
            # class LatLong:
            #     lat: Optional[float]
            #     long: Optional[float]
            # then self._from_prim(None) == LatLong(None, None)
            # Using self.primitive_type, rather than self._rich_type, as the primitive type
            # might not be registered on the converter for custom classes
            try:
                x = unstructure_rich_to_primitive(val)
                y = structure_primitive_to_rich(x, cast(Type[_TRich], self.primitive_type))
                return cast(Optional[_TPrim], y)
            except (TypeError, ValueError) as e:
                raise TypeError(f"Could not convert '{val}' to `{self.primitive_type}`") from e
        return self._encoder(val)

    def _from_prim(self, val: Optional[Union[_TPrim, _TRich]]) -> Optional[_TRich]:
        if isinstance(self._rich_type, _UnknownType):
            raise ValueError(
                "Rich types cannot be used as the FeatureConverter was created without providing a `rich_type`"
            )
        if val is None:
            # Structuring null values to the primitive type to ensure that a singular null for an entire struct
            # is propagated to individual struct fields -- e.g.
            # class LatLong:
            #     lat: Optional[float]
            #     long: Optional[float]
            # then self._from_prim(None) == LatLong(None, None)
            # Using self.primitive_type, rather than self._rich_type, as the primitive type
            # might not be registered on the converter for custom classes
            try:
                val = structure_primitive_to_rich(cast(_TPrim, val), cast(Type[_TRich], self.primitive_type))
            except (TypeError, ValueError) as e:
                raise TypeError(f"Could not convert '{val}' to `{self.primitive_type}`") from e

        if self._decoder is None:
            try:
                return structure_primitive_to_rich(cast(_TPrim, val), self._rich_type)
            except (TypeError, ValueError) as e:
                raise TypeError(f"Could not convert '{val}' to `{self._rich_type}`") from e
        if isinstance(val, self._rich_type):
            return cast(_TRich, val)
        return self._decoder(cast(_TPrim, val))

    def from_rich_to_pyarrow(self, values: Sequence[Optional[_TRich]], /) -> pa.Array:
        prim_values = [self.from_rich_to_primitive(x) for x in values]
        return self.from_primitive_to_pyarrow(prim_values)

    def from_rich_to_primitive(self, value: Optional[_TRich]) -> Optional[_TPrim]:
        # Ensure that the rich value is indeed the rich type
        # For example, if a string is passed in for a datetime value, convert it into a datetime
        value = self.from_primitive_to_rich(cast(_TPrim, value))
        return self._to_primitive(value)

    def from_rich_to_json(self, value: Optional[_TRich]) -> TJSON:
        prim_val = self.from_rich_to_primitive(value)
        return self.from_primitive_to_json(prim_val)

    def from_pyarrow_to_rich(self, values: pa.Array, /) -> Sequence[Optional[_TRich]]:
        return [self.from_primitive_to_rich(x) for x in values.to_pylist()]

    def from_pyarrow_to_json(self, values: pa.Array) -> Sequence[TJSON]:
        return [self.from_primitive_to_json(x) for x in self.from_pyarrow_to_primitive(values)]

    def from_pyarrow_to_primitive(self, values: pa.Array) -> Sequence[Optional[_TPrim]]:
        return values.to_pylist()

    def from_primitive_to_rich(self, value: Optional[_TPrim]) -> Optional[_TRich]:
        return self._from_prim(value)

    def from_primitive_to_pyarrow(self, value: Sequence[Optional[_TPrim]]) -> pa.Array:
        x = pa.array(value, type=self._pyarrow_dtype)
        return x

    def from_primitive_to_json(self, value: Optional[_TPrim]) -> TJSON:
        return unstructure_primitive_to_json(value)

    def from_json_to_rich(self, value: TJSON) -> Optional[_TRich]:
        prim_val = self.from_json_to_primitive(value)
        return self.from_primitive_to_rich(prim_val)

    def from_json_to_pyarrow(self, values: Sequence[TJSON]) -> pa.Array:
        primitive_vals = [cast(_TPrim, self.from_json_to_primitive(x)) for x in values]
        return self.from_primitive_to_pyarrow(primitive_vals)

    def from_json_to_primitive(self, value: TJSON) -> Optional[_TPrim]:
        try:
            return cast(Optional[_TPrim], structure_json_to_primitive(value, self._primitive_type))
        except (ValueError, TypeError) as e:
            raise TypeError(f"Could not convert '{value}' to `{self._primitive_type}`") from e

    @property
    def pyarrow_dtype(self):
        return self._pyarrow_dtype

    @property
    def rich_type(self) -> Type[_TRich]:
        if isinstance(self._rich_type, _UnknownType):
            raise ValueError(
                "Rich types cannot be used as the FeatureConverter was created without providing a `rich_type`"
            )
        return self._rich_type

    @property
    def primitive_type(self) -> Type[TPrimitive]:
        return self._primitive_type

    @functools.cached_property
    def polars_dtype(self) -> Union[pl.DataType, Type[pl.DataType]]:
        return pyarrow_to_polars(self.pyarrow_dtype, self._name)

    @property
    def serialized_dtype(self):
        return self._serialized_dtype



from typing import Any, Type, TypeVar, cast

from typing_extensions import Annotated

from chalk.features.feature_wrapper import unwrap_feature

T = TypeVar("T")


class PrimaryMeta(type):
    def __getitem__(self, item: Type[T]) -> Type[T]:
        return cast(Type[T], Annotated[item, "__chalk_primary__"])


Primary = PrimaryMeta("Primary", (object,), {})


def is_primary(f: Any) -> bool:
    """
    Determine whether a feature is a primary key
    :param f: A feature (ie. User.email)
    :return: True if the feature is primary and False otherwise
    """
    return unwrap_feature(f).primary



from chalk.utils.string import removesuffix


def test_remove_suffix():
    assert removesuffix("joe's waffles", "waffles") == "joe's "
    assert removesuffix("joe's waffles", "bananas") == "joe's waffles"



import pytest

from chalk.features import FeatureSetBase, features, has_one, unwrap_feature
from chalk.parsed.json_conversions import convert_type_to_gql


def test_converting_filters():
    @features
    class InvalidFeatureClass2:
        id: int

    @features
    class InvalidFeatureClass:
        id: int

        relationship_id: int
        relationship: InvalidFeatureClass2 = has_one(
            lambda: InvalidFeatureClass.relationship == InvalidFeatureClass2.id
        )
        relationship2: InvalidFeatureClass2 = has_one(
            lambda: InvalidFeatureClass.relationship_id == InvalidFeatureClass2.id
        )

    # should fail -- referencing a has one in the filter
    with pytest.raises(AssertionError):
        convert_type_to_gql(unwrap_feature(InvalidFeatureClass.relationship))

    # should be fine
    convert_type_to_gql(unwrap_feature(InvalidFeatureClass.relationship2))

    del FeatureSetBase.registry["invalid_feature_class"]
    del FeatureSetBase.registry["invalid_feature_class2"]



import pytest
from pydantic import BaseModel

from chalk.features import DataFrame, features
from chalk.features.resolver import StreamResolver
from chalk.sql._internal.sql_file_resolver import SQLStringResult, get_sql_file_resolver
from chalk.streams import KafkaSource, Windowed, stream, windowed

try:
    import duckdb
except ImportError:
    duckdb = None


try:
    import polars as pl
except ImportError:
    pl = None


@features
class StoreFeatures:
    id: str
    purchases: Windowed[float] = windowed("10m", "20m")


class KafkaMessage(BaseModel):
    purchase_id: str
    store_id: str
    amount: float


source = KafkaSource(bootstrap_server="server", topic="topic", name="kafka_test_source")


@stream(source=source)
def fn(messages: DataFrame[KafkaMessage]) -> DataFrame[StoreFeatures.id, StoreFeatures.purchases]:
    return f"""
        select store_id as id, sum(amount) as purchases
        from {messages}
        group by 1
    """


@pytest.mark.skipif(duckdb is None or pl is None, reason="duckdb is not installed or polars is not installed")
def test_runs_sql():
    assert pl is not None
    expected = DataFrame(
        pl.DataFrame(
            {
                "store_features.id": ["store1", "store2"],
                "store_features.purchases": [10, 5],
            },
        )
    )
    actual = fn(
        DataFrame(
            pl.DataFrame(
                {
                    "store_id": ["store1", "store2", "store1"],
                    "amount": [4, 5, 6],
                }
            ),
            pydantic_model=KafkaMessage,
        )
    )

    # polars.testing.assert_frame_equal(actual, expected)
    assert expected.to_pyarrow() == actual.to_pyarrow()


@pytest.mark.skipif(duckdb is None or pl is None, reason="duckdb is not installed or polars is not installed")
def test_sql_file_stream_resolver():
    sources = [source]
    sql_string = """-- source: kafka_test_source
        -- resolves: store_features
        -- type: streaming
        select store_id as id, sum(amount) as purchases
        from topic
        group by 1
    """
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert not resolver_result.errors
    assert resolver_result.fields == {"id": "store_features.id", "purchases": "store_features.purchases"}
    res = resolver_result.resolver
    assert (
        res.sql_query
        == """select store_id as id, sum(amount) as purchases
        from topic
        group by 1"""
    )
    assert res.source == source


@pytest.mark.skipif(duckdb is None or pl is None, reason="duckdb is not installed or polars is not installed")
def test_sql_file_stream_resolver_kafka():
    sources = [source]
    sql_string = """-- source: kafka
        -- resolves: store_features
        -- type: streaming
        select store_id as id, sum(amount) as purchases
        from topic
        group by 1
    """
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert not resolver_result.errors
    assert resolver_result.fields == {"id": "store_features.id", "purchases": "store_features.purchases"}
    res = resolver_result.resolver
    assert isinstance(res, StreamResolver)
    assert (
        res.sql_query
        == """select store_id as id, sum(amount) as purchases
        from topic
        group by 1"""
    )



from __future__ import annotations

import warnings
from datetime import datetime
from typing import TYPE_CHECKING, Any, Type, Union, cast

import cattrs
import pytz
from dateutil import parser

if TYPE_CHECKING:
    import polars as pl

    from chalk.features import Feature


def is_numeric(dtype: Union[pl.DataType, Type[pl.DataType]]) -> bool:
    warnings.warn(DeprecationWarning("This method is deprecated"))
    # This function is used only by the engine
    import polars as pl

    return dtype in (
        pl.Float32,
        pl.Float64,
        pl.Int16,
        pl.Int8,
        pl.Int32,
        pl.Int64,
        pl.UInt8,
        pl.UInt16,
        pl.UInt32,
        pl.UInt64,
    )


class FeatureCodec:
    def __init__(self) -> None:
        self.converter = cattrs.Converter()
        self.converter.register_structure_hook(datetime, lambda v, _: parser.isoparse(v))
        self.converter.register_unstructure_hook(
            datetime,
            lambda v: (
                cast(datetime, v) if cast(datetime, v).tzinfo else pytz.utc.localize(cast(datetime, v))
            ).isoformat(),
        )

    def encode(
        self,
        feature: Feature,
        value: Any,
    ):
        # This method is only used by the engine
        warnings.warn(
            DeprecationWarning(
                "FeatureCodec.encode is deprecated. Instead, use `feature.converter.from_rich_to_json(...)`"
            )
        )
        return feature.converter.from_rich_to_json(value)

    def encode_fqn(self, fqn: str, value: Any):
        # This method is only used by the engine
        from chalk.features import Feature

        return self.encode(Feature.from_root_fqn(fqn), value)

    def decode(
        self,
        feature: Feature,
        value: Any,
    ):
        # This method is only used by the engine
        warnings.warn(
            DeprecationWarning(
                "FeatureCodec.decode is deprecated. Instead, use `feature.converter.from_json_to_rich(...)`"
            )
        )

        return feature.converter.from_json_to_rich(value)

    def get_polars_dtype(self, fqn: str) -> Union[pl.DataType, Type[pl.DataType]]:
        # This function is used only by the engine
        warnings.warn(
            DeprecationWarning(
                (
                    "FeatureCodec.get_polars_dtype is deprecated. Instead, use switch to PyArrow dtypes. "
                    "The PyArrow dtype can be found at `feature.converter.pyarrow_dtype`"
                )
            )
        )
        import polars as pl

        from chalk.features import Feature

        feature = Feature.from_root_fqn(fqn)

        if feature.is_has_one or feature.is_has_many:
            warnings.warn("FeatureCodec.get_polars_dtype should not be called on has-one / has-many features")
            return pl.Object
        else:
            return feature.converter.polars_dtype


FEATURE_CODEC = FeatureCodec()
